import pytest
import numpy as np
from deepview.videostream import Frame
from pathlib import Path
from random import randint
from warnings import warn


def test_frame():
    """
        Basic test demonstrating we can create two frames which share memory.
        In this case frame 1 will allocate from the default space: dma_heap if
        available, otherwise shared memory.  We then create frame 2 using the
        handle from frame 1.  We demonstrate a few examples where the memory
        is shared, including zero-copy memoryviews with Numpy.
    """
    frame = Frame(640, 480, fourcc='RGB3')
    assert frame.width == 640
    assert frame.height == 480
    assert frame.fourcc == 'RGB3'
    assert frame.path is None

    with pytest.raises(Exception):
        frame.handle

    frame.alloc()
    assert frame.size == 640 * 480 * 3

    v1 = frame.map()
    assert len(v1) == 640 * 480 * 3

    a1 = np.frombuffer(v1, dtype=np.uint8)
    a1[:] = np.random.randint(0, 255, len(v1), dtype=np.uint8)
    assert (a1 == v1).all()
    frame2 = Frame(640, 480, fourcc='RGB3')
    frame2.attach(frame.handle)
    v2 = frame2.map()
    assert (a1 == v2).all()
    for i in range(len(v2)):
        v2[i] = randint(0, 255)
    assert v1 == v2

    try:
        assert frame.paddr != 0
    except Exception as err:
        warn('Physical Address: %s' % err, UserWarning)

    frame2.unmap()
    frame2.unalloc()

    frame.unmap()
    frame.unalloc()

    with pytest.raises(Exception):
        frame.handle


def test_attach_file():
    """
        This test demonstrates how we can create a frame in our filesystem then
        associated the file descriptor to a videostream frame to use normally.
    """
    frame = Frame(640, 480, fourcc='RGB3')
    expect = np.random.randint(
        0, 255, frame.stride * frame.height, dtype=np.uint8)
    # Note the mode 'w+b' is important here as without the ability to  write
    # 'w' AND read 'w+' the mapping will fail.  The 'b' represents binary.
    with open('/tmp/test_frame.vsl', 'w+b') as f:
        f.truncate(frame.stride * frame.height)
        frame.attach(f.fileno())
    v = frame.map()
    v[:] = expect
    del v
    frame.unmap()
    out = np.fromfile('/tmp/test_frame.vsl', dtype=np.uint8)
    assert (expect == out).all()


def test_bad_attach():
    """
        Tests various invalid file descriptor attachments to ensure they throw
        errors and don't lead to further issues, such as crashes.
    """
    frame = Frame(640, 480, fourcc='RGB3')
    with pytest.raises(Exception):
        frame.attach(-1)

    # Assuming fd are allocated sequentially, currently the case, and picking a
    # large enough fd that it shouldn't be open currently.
    with pytest.raises(Exception):
        frame.attach(9000)

    # The 0 fd should typically be stdin, attach doesn't fail but mmap does.
    frame.attach(0)
    with pytest.raises(Exception):
        frame.map()


@pytest.mark.parametrize("fourcc,stride", [
    ('RGB3', 3), ('RGBX', 4), ('RGBA', 4),
    ('BGR3', 3), ('BGRX', 4), ('BGRA', 4),
    ('YUYV', 2), ('UYVY', 2), ('YVYU', 2), ('VYUY', 2), ('YUY2', 2),
    ('I420', 1.5), ('NV12', 1.5), ('YV12', 1.5), ('NV12', 1.5),
])
def test_fourcc(fourcc, stride):
    """
        Test we can create frames for some known fourcc codes and that others
        will raise exceptions for the unsupported formats.  We also test some
        assumptions about frame sizes and strides for various fourcc codes.
    """
    frame = Frame(4, 3, fourcc)
    assert frame.stride == stride * 4
    frame.alloc()
    assert frame.size == stride * 4 * 3


@pytest.mark.parametrize("fourcc", ['GRB', '    ', '', None])
def test_bad_fourcc(fourcc):
    """
        Test for some expected errors when using invalid or unsupported fourcc.
    """
    with pytest.raises(Exception):
        Frame(4, 3, fourcc)


def test_nodma():
    """
        Test verifies that attempting to allocate from dma_heap fails when it
        is not available.  We test a dma_heap which is expected to never exist.
    """
    frame = Frame(640, 480, fourcc='RGB3')
    with pytest.raises(Exception):
        frame.alloc('/dev/dma_heap/invalid')


def test_invalid_shm_name():
    """
        When not using dma_heap, VSL uses POSIX Shared Memory for buffers.  It
        requires names to start with a / but not have additional / in the name.

        These are not actual filesystem objects, so giving a filesystem path
        such as /tmp/invalid should raise an error.
    """
    frame = Frame(640, 480, fourcc='RGB3')
    with pytest.raises(Exception):
        frame.alloc('/tmp/invalid')


def test_shm_fill_benchmark(benchmark):
    """
        The Shared Memory fill benchmark.
    """
    frame = Frame(640, 480, fourcc='RGB3')
    frame.alloc('/VSL_BENCH_SHM')
    v1 = np.frombuffer(frame.map(), dtype=np.uint8).reshape(480, 640, 3)
    assert v1.size == 640 * 480 * 3

    def fill():
        v1[:] = np.ones((480, 640, 3), dtype=np.uint8)
    benchmark(fill)


def test_dma_fill_benchmark(benchmark):
    """
        The DMA fill benchmark.
    """
    dma_heap_path = Path('/dev/dma_heap')
    if not dma_heap_path.exists():
        pytest.skip(
            "DMA heap not available - test requires /dev/dma_heap "
            "(embedded target only)"
        )

    # Verify DMA heap is actually usable by attempting allocation
    try:
        test_frame = Frame(64, 64, fourcc='RGB3')
        test_frame.alloc()
        if not test_frame.path.startswith('/dev/dma_heap'):
            pytest.skip(
                "DMA heap exists but allocation fell back to shared "
                "memory - check /dev/dma_heap permissions (may need "
                "video group membership)"
            )
        del test_frame
    except Exception as e:
        pytest.skip(
            f"DMA heap exists but cannot allocate: {e}"
        )

    frame = Frame(640, 480, fourcc='RGB3')
    frame.alloc()
    if not frame.path.startswith('/dev/dma_heap'):
        pytest.skip(
            f"Frame allocation unexpectedly fell back to "
            f"{frame.path} - DMA heap may be out of memory"
        )
    v1 = np.frombuffer(frame.map(), dtype=np.uint8).reshape(480, 640, 3)
    assert v1.size == 640 * 480 * 3

    def fill():
        v1[:] = np.ones((480, 640, 3), dtype=np.uint8)
    benchmark(fill)


def test_dma_fill_remap_benchmark(benchmark):
    """
        The DMA fill benchmark, but with added unmap/map in the loop which
        will trigger cache flush/refresh.  This benchmark will be slower
        because of the cache synchronization and represents the cost of DMA
        buffers when sharing with hardware.
    """
    dma_heap_path = Path('/dev/dma_heap')
    if not dma_heap_path.exists():
        pytest.skip(
            "DMA heap not available - test requires /dev/dma_heap "
            "(embedded target only)"
        )

    # Verify DMA heap is actually usable by attempting allocation
    try:
        test_frame = Frame(64, 64, fourcc='RGB3')
        test_frame.alloc()
        if not test_frame.path.startswith('/dev/dma_heap'):
            pytest.skip(
                "DMA heap exists but allocation fell back to shared "
                "memory - check /dev/dma_heap permissions (may need "
                "video group membership)"
            )
        del test_frame
    except Exception as e:
        pytest.skip(
            f"DMA heap exists but cannot allocate: {e}"
        )

    frame = Frame(640, 480, fourcc='RGB3')
    frame.alloc()
    if not frame.path.startswith('/dev/dma_heap'):
        pytest.skip(
            f"Frame allocation unexpectedly fell back to "
            f"{frame.path} - DMA heap may be out of memory"
        )

    def fill():
        v1 = np.frombuffer(frame.map(), dtype=np.uint8).reshape(480, 640, 3)
        v1[:] = np.ones((480, 640, 3), dtype=np.uint8)
        del v1  # Explicitly remove v1 before frame.unmap()
        frame.unmap()
    benchmark(fill)


def test_dma_fill_sync_benchmark(benchmark):
    """
        The DMA fill benchmark, but with added sync in the loop to provide
        more fine-trained synchronization than forcing map/unmap.

        This benchmark will be slower because of the cache synchronization
        and represents the cost of DMA buffers when sharing with hardware.
    """
    dma_heap_path = Path('/dev/dma_heap')
    if not dma_heap_path.exists():
        pytest.skip(
            "DMA heap not available - test requires /dev/dma_heap "
            "(embedded target only)"
        )

    # Verify DMA heap is actually usable by attempting allocation
    try:
        test_frame = Frame(64, 64, fourcc='RGB3')
        test_frame.alloc()
        if not test_frame.path.startswith('/dev/dma_heap'):
            pytest.skip(
                "DMA heap exists but allocation fell back to shared "
                "memory - check /dev/dma_heap permissions (may need "
                "video group membership)"
            )
        del test_frame
    except Exception as e:
        pytest.skip(
            f"DMA heap exists but cannot allocate: {e}"
        )

    frame = Frame(640, 480, fourcc='RGB3')
    frame.alloc()
    if not frame.path.startswith('/dev/dma_heap'):
        pytest.skip(
            f"Frame allocation unexpectedly fell back to "
            f"{frame.path} - DMA heap may be out of memory"
        )
    v1 = np.frombuffer(frame.map(), dtype=np.uint8).reshape(480, 640, 3)
    frame.sync(False, Frame.Sync.RW)  # Disable sync after map

    def fill():
        frame.sync(True, Frame.Sync.WRONLY)
        v1[:] = np.ones((480, 640, 3), dtype=np.uint8)
        frame.sync(False, Frame.Sync.WRONLY)
    benchmark(fill)


def test_numpy_fill_benchmark(benchmark):
    """
        The reference benchmark using standard Numpy buffers.  This benchmark
        should not be measurably faster or slower than the shared memory or
        dma_heap benchmarks.
    """
    v1 = np.zeros((480, 640, 3), dtype=np.uint8)
    assert v1.size == 640 * 480 * 3

    def fill():
        v1[:] = np.ones((480, 640, 3), dtype=np.uint8)
    benchmark(fill)



import pytest
from deepview.videostream import Client, Host, Frame
from pathlib import Path


def test_connect():
    """
    Test basic connection support between host and client.  Tests need to be
    careful to ensure host is ready for client otherwise the client will block
    waiting for the host to be ready, which will never happen.  Later tests add
    threading to the mix to better orchestrate the communications tests.
    """
    h = Host('/tmp/test_connect.vsl')
    assert Path(h.path).exists()
    c = Client(h.path)
    assert c.path == h.path

    h.process()
    assert h.clients_count == 1

    c.disconnect()
    h.process()
    assert h.clients_count == 0


def test_post():
    """
    """
    h = Host('/tmp/test_post.vsl')
    assert Path(h.path).exists()
    c = Client(h.path)
    assert c.path == h.path

    h.process()
    assert h.clients_count == 1

    f1 = Frame(640, 480, fourcc='RGB3')
    f1.alloc()
    assert f1.size == 640 * 480 * 3
    f1_handle = f1.handle

    h.post(f1)
    assert f1._ptr is None
    with pytest.raises(Exception):
        f1.handle
        f1.width
        f1.height
        f1.fourcc

    f2 = c.frame_wait()
    assert f2 is not None
    assert f2._ptr is not None
    assert f2.width == 640
    assert f2.height == 480
    assert f1_handle != f2.handle



import pytest
from deepview.videostream import Host
from pathlib import Path
from socket import socket, AF_UNIX, SOCK_SEQPACKET
from time import process_time
from warnings import warn


def test_host():
    """
    Tests basic host boilerplate such as creating the host creates the socket
    and proper error handling for bad or missing parameters.
    """
    with pytest.raises(Exception):
        h = Host()
    with pytest.raises(Exception):
        h = Host(None)
    with pytest.raises(Exception):
        h = Host('')

    p = Path('/tmp/test_host.vsl')
    assert not p.exists()
    h = Host(p)
    assert p.exists()
    assert h.path == p.as_posix()
    assert h.clients_count == 0

    s = socket(fileno=h.listener)
    assert s.family == AF_UNIX
    assert s.type == SOCK_SEQPACKET
    assert s.getsockname() == p.as_posix()

    # Poll with a one second timeout and assert that we
    # elapsed at least one second but less than two seconds.
    start = process_time()
    assert h.poll(1.0) == 0
    elapsed = process_time() - start
    # FIXME: function is returning after about 100µs need to investigate.
    assert elapsed < 2.0
    if elapsed < 1.0:
        warn('Host poll expected runtime over 1000ms but took %.03fms' %
             (elapsed * 1000.0))

    del h
    assert not p.exists()



# SPDX-License-Identifier: Apache-2.0
# Copyright Ⓒ 2025 Au-Zone Technologies. All Rights Reserved.

from ctypes import \
    byref, c_int, c_int64, c_size_t, c_void_p, py_object, pythonapi
from deepview.videostream.library import lib, errmsg
from enum import IntEnum
from pathlib import Path
from typing import Optional, Union
from typeguard import typechecked
from .frame import Frame


@typechecked
class Host:
    """
    The VideoStream host allows sharing of frames to clients.  Frames created
    by the host can be posted and connected clients can receive them.  The
    sharing is done through a UNIX Domain Socket but only the frame information
    is transmitted with the descriptor for the underlying buffer being shared
    through the anciliary channel.  This allows for zero-copy frame sharing
    between the host and clients.
    """

    def __init__(self, path: Union[str, Path]):
        if isinstance(path, Path):
            path = path.as_posix()
        self._ptr = lib.vsl_host_init(path.encode('utf-8'))
        if self._ptr is None:
            raise RuntimeError('Failed to create host: %s' % errmsg())

    def __del__(self):
        if hasattr(self, '_ptr') and self._ptr is not None:
            lib.vsl_host_release(self._ptr)
            self._ptr = None

    @property
    def path(self) -> str:
        p = lib.vsl_host_path(self._ptr)
        if p is None:
            raise RuntimeError('Failed to read host path: %s' % errmsg())
        return p.decode('utf-8')

    @property
    def listener(self) -> int:
        s = c_int()
        ret = lib.vsl_host_sockets(self._ptr, c_size_t(1), byref(s), None)
        if ret == -1:
            raise RuntimeError(
                'Failed to get listener socket: %s' % errmsg())
        return s.value

    @property
    def clients_count(self) -> int:
        """
        Returns the number of clients connected to the host.
        """
        n = c_size_t()
        ret = lib.vsl_host_sockets(self._ptr, 0, None, byref(n))
        if ret == -1:
            raise RuntimeError(
                'Failed to get client sockets: %s' % errmsg())
        # Account for the listener socket which will always be present.
        return n.value - 1

    def poll(self, timeout: float = 0.0) -> int:
        """
        Polls all host sockets, including the listener, for readiness.
        """
        t = c_int64(int(timeout * 1000))
        ret = lib.vsl_host_poll(self._ptr, t)
        if ret == -1:
            raise RuntimeError('Failed to poll host socket: %s' % errmsg())
        return ret

    def process(self) -> None:
        """
        Process host sockets, accepting new clients and servicing connected
        clients.
        """
        ret = lib.vsl_host_process(self._ptr)
        if ret == -1:
            raise RuntimeError('Failed to process host: %s' % errmsg())

    def post(self,
             frame: Frame,
             expires: int = 0,
             duration: int = 0,
             pts: int = 0,
             dts: int = 0) -> None:
        ret = lib.vsl_host_post(self._ptr, frame._ptr,
                                expires, duration, pts, dts)
        if ret == -1:
            raise RuntimeError('Failed to post frame: %s' % errmsg())
        frame._ptr = None
        del frame



# SPDX-License-Identifier: Apache-2.0
# Copyright Ⓒ 2025 Au-Zone Technologies. All Rights Reserved.

from ctypes import byref, c_int, c_size_t, c_void_p, py_object, pythonapi
from deepview.videostream.library import lib
from enum import IntEnum
from typing import Optional
from typeguard import typechecked


@typechecked
class Frame:
    """
    VideoStream Frame class manages frames.
    """

    @typechecked
    class Sync(IntEnum):
        """
        When manually controlling DMA sync these definitions are used to state
        whether the sync is for read-only, write-only, or read-write operation.
        """
        RDONLY = 0
        WRONLY = 1
        RW = 2

    def __init__(self, width, height, fourcc, stride=0, ptr=None):
        """
        Constructs the frame for the provided width, height, and fourcc.
        """
        if ptr is not None:
            self._ptr = ptr
            return
        if type(fourcc) == str:
            fourcc_ = lib.vsl_fourcc_from_string(fourcc.encode('utf-8'))
            if fourcc_ == 0:
                raise RuntimeError('Invalid fourcc %s' % fourcc)
            fourcc = fourcc_
        self._ptr = lib.vsl_frame_init(
            width, height, stride, fourcc, None, None)
        if self._ptr is None:
            raise RuntimeError('Failed to create Frame')

    def __del__(self):
        if hasattr(self, '_ptr') and self._ptr is not None:
            lib.vsl_frame_release(self._ptr)
            self._ptr = None

    @property
    def width(self) -> int:
        """
        Returns the width of the frame in pixels.
        """
        return lib.vsl_frame_width(self._ptr)

    @property
    def height(self) -> int:
        """
        Returns the height of the frame in pixels.
        """
        return lib.vsl_frame_height(self._ptr)

    @property
    def stride(self) -> int:
        """
        Returns the number of bytes to the next row.  This value is not the
        width but rather the width * pixel_size + padding; where pixel_size
        refers to the number of bytes for a major pixel, and padding is extra
        area before the next row.

        Major pixel has varying meaning.  For RGB type images a pixel is the
        combination of R,G,B, and possibly A or X (X means ignored).  An RGB
        image could also be called RGB888 which refers to 3 8-bit pixels. For
        YUV type images it gets a little more complicated.  For packed formats
        such as YUYV the pixel stride is 2 (16bits per pixel) as the pixels are
        an alternating pattern of YU/YV, this also implies the resolution must
        be a factor of 2.  For planar formats such as I420 or NV12 the YUV
        planes are separated, differently.  I420 has a plane for Y and a plane
        for UV (packed).  NV12 has a plane for each of Y, U, and V.  There is
        also some discrepancy in interpretation between vendors, our choice is
        based on the underlying hardware/software/drivers used by videostream.

        Padding is commonly used for cropping, by adjusting the starting pixel
        you can choose the x,y starting coordinate from the left,top along with
        stride to overshoot the width of the image to align with the next x
        coordinate.
        """
        return lib.vsl_frame_stride(self._ptr)

    @property
    def fourcc(self) -> str:
        """
        Returns the FOURCC code as a string.
        """
        code = lib.vsl_frame_fourcc(self._ptr)
        return '%c%c%c%c' % (
            chr(code & 0xFF),
            chr((code >> 8) & 0xFF),
            chr((code >> 16) & 0xFF),
            chr((code >> 24) & 0xFF))

    @property
    def size(self) -> int:
        """
        Returns the size of the underlying framebuffer.
        """
        sz = lib.vsl_frame_size(self._ptr)
        if sz <= 0:
            raise RuntimeError('Frame has unknown size')
        return sz

    @property
    def paddr(self) -> int:
        """
        Returns the physical address of the underlying buffer.

        Note: this is only available on certain embedded platforms.
        """
        p = lib.vsl_frame_paddr(self._ptr)
        if p == -1:
            raise RuntimeError('Frame cannot produce physical address')
        return p

    @property
    def path(self) -> Optional[str]:
        """
        Returns the path of the underlying frame buffer, if it is known.
        """
        s = lib.vsl_frame_path(self._ptr)
        return None if s is None else s.decode('UTF-8')

    @property
    def handle(self) -> int:
        """
        Returns the file descriptor for the underlying frame buffer.
        """
        fd = lib.vsl_frame_handle(self._ptr)
        if fd == -1:
            raise RuntimeError('Frame does not have a known file handle')
        return fd

    def map(self) -> memoryview:
        """
        Maps the buffer and returns a memoryview of the object.

        DANGEROUS: Note that you are responsible for calling unmap after you
        are done with the memoryview and MUST ensure to never access the buffer
        once it has been unmapped.  If you cannot guarantee the safety of your
        code then you must instead make a copy of the memoryview then dispose
        and do the unmap operation.
        """
        siz = c_size_t()
        ptr = lib.vsl_frame_mmap(self._ptr, byref(siz))
        if ptr is None:
            raise RuntimeError('Failed to map frame')
        from_memory = pythonapi.PyMemoryView_FromMemory
        from_memory.argtypes = (c_void_p, c_int, c_int)
        from_memory.restype = py_object
        return from_memory(ptr, siz.value, 0x200)

    def unmap(self) -> None:
        """
        Release the frame mapping.  Once you call unmap the previous memoryview
        returned by map() is now invalid!
        """
        lib.vsl_frame_munmap(self._ptr)

    def sync(self, enable: bool, mode: Sync) -> None:
        """
        Cache synchronization session control for when using DMA-backed
        buffers.

        This happens automatically on mmap/munmap but the API is also available
        for cases where the frame is updated in-place during a mapping.
        """
        ret = lib.vsl_frame_sync(self._ptr, 1 if enable else 0, mode.value)
        if ret == -1:
            raise RuntimeError('Failed to sync frame')

    def alloc(self, path=None):
        if path is not None:
            path = path.encode('UTF-8')
        ret = lib.vsl_frame_alloc(self._ptr, path)
        if ret == -1:
            raise RuntimeError('Failed to allocate frame')

    def unalloc(self):
        lib.vsl_frame_unalloc(self._ptr)

    def attach(self, fd, size=0, offset=0):
        ret = lib.vsl_frame_attach(self._ptr, fd, size, offset)
        if ret == -1:
            raise RuntimeError('Failed to attach buffer to frame')



# SPDX-License-Identifier: Apache-2.0
# Copyright Ⓒ 2025 Au-Zone Technologies. All Rights Reserved.

from .library import lib, version
from .frame import Frame
from .client import Client
from .host import Host



import pytest
from deepview.videostream import Client


def test_client():
    """
    Testing some basic client setup code.  The main tests are in test_ipc which
    require both the host and client to be setup for communications.
    """
    with pytest.raises(Exception):
        Client()
    with pytest.raises(Exception):
        Client(None)
    with pytest.raises(Exception):
        Client('')



# SPDX-License-Identifier: Apache-2.0
# Copyright Ⓒ 2025 Au-Zone Technologies. All Rights Reserved.

from ctypes import \
    CDLL, POINTER, get_errno, \
    c_char_p, c_void_p, c_size_t, c_ssize_t, \
    c_int, c_int64, c_uint32, c_bool, c_float
from os import environ, strerror
from os.path import isdir, join
from typing import Optional
from typeguard import typechecked


@typechecked
def version() -> str:
    """
    Version of the underlying videostream.ext library, the .ext varies across
    platforms.

        * Windows: ``VideoStream.dll``
        * Linux: ``libvideostream.so``
        * MacOS: ``libvideostream.dylib``

    Returns:
        version string as major.minor.patch-extra format.
    """
    string = lib.vsl_version()
    return c_char_p(string).value.decode('utf-8')


@typechecked
def timestamp() -> int:
    """
    Returns the current VSL timestamp in nanoseconds.
    """
    return lib.vsl_timestamp()


@typechecked
def errmsg() -> Optional[str]:
    """
    Returns the error message for the current system error code (errno).
    """
    return strerror(get_errno())


@typechecked
def load_library(libname: Optional[str] = None) -> CDLL:
    """
    Internal function used to load and configure the VideoStream library. This
    function should not be called directly but instead is called automatically
    when the deepview.videostream library is first loaded.

    The environment variable ``VIDEOSTREAM_LIBRARY`` can be used to point to
    the location of ``videostream.dll/dylib/so`` for cases where it cannot be
    found.

    Note:
        The library is not part of the Python package but is installed
        separately, typically as part of DeepView VisionPack installations.

    Returns:
        A ctypes.CDLL object containing the VideoStream library.

    Raises:
        :py:class:`EnvironmentError`: if the VideoStream library cannot be
        located.
    """
    if 'VIDEOSTREAM_LIBRARY' in environ:
        libname = environ['VIDEOSTREAM_LIBRARY']

    if libname is not None:
        if isdir(libname) and libname.endswith('.framework'):
            return CDLL(
                join(libname, 'Versions', 'Current', 'VideoStream'))
        else:
            return CDLL(libname)
    else:
        try:
            return CDLL('VideoStream.dll')
        except OSError:
            pass

        try:
            return CDLL('./VideoStream.dll')
        except OSError:
            pass

        try:
            return CDLL('libvideostream.so')
        except OSError:
            pass

        try:
            return CDLL('libvideostream.dylib')
        except OSError:
            pass

    raise EnvironmentError(
        'Unable to load the VideoStream library.  Try setting the environment \
         variable VIDEOSTREAM_LIBRARY to the VideoStream library.')


@typechecked
def load_symbols(lib: CDLL):
    """
    Loads the symbols from the VideoStream library into the `lib` object to be
    used by the various Python API.

    These symbols are documented in videostream.h and the VideoStream Library
    User Manual.

    Args:
        lib: Library object returned from :func:`load_library()`
    """

    # Core Library API
    lib.vsl_version.argtypes = []
    lib.vsl_version.restype = c_char_p

    lib.vsl_timestamp.argtypes = []
    lib.vsl_timestamp.restype = c_int64

    lib.vsl_fourcc_from_string.argtypes = [c_char_p]
    lib.vsl_fourcc_from_string.restype = c_uint32

    # Frame API
    lib.vsl_frame_init.argtypes = [
        c_uint32, c_uint32, c_uint32, c_uint32, c_void_p, c_void_p]
    lib.vsl_frame_init.restype = c_void_p

    lib.vsl_frame_release.argtypes = [c_void_p]
    lib.vsl_frame_release.restype = None

    lib.vsl_frame_width.argtypes = [c_void_p]
    lib.vsl_frame_width.restype = c_int

    lib.vsl_frame_height.argtypes = [c_void_p]
    lib.vsl_frame_height.restype = c_int

    lib.vsl_frame_stride.argtypes = [c_void_p]
    lib.vsl_frame_stride.restype = c_int

    lib.vsl_frame_size.argtypes = [c_void_p]
    lib.vsl_frame_size.restype = c_int

    lib.vsl_frame_fourcc.argtypes = [c_void_p]
    lib.vsl_frame_fourcc.restype = c_uint32

    lib.vsl_frame_path.argtypes = [c_void_p]
    lib.vsl_frame_path.restype = c_char_p

    lib.vsl_frame_handle.argtypes = [c_void_p]
    lib.vsl_frame_handle.restype = c_int

    lib.vsl_frame_paddr.argtypes = [c_void_p]
    lib.vsl_frame_paddr.restype = c_ssize_t

    lib.vsl_frame_attach.argtypes = [c_void_p, c_int, c_size_t, c_size_t]
    lib.vsl_frame_attach.restype = c_int

    lib.vsl_frame_alloc.argtypes = [c_void_p, c_char_p]
    lib.vsl_frame_alloc.restype = c_int

    lib.vsl_frame_unalloc.argtypes = [c_void_p]
    lib.vsl_frame_unalloc.restype = None

    lib.vsl_frame_mmap.argtypes = [c_void_p, POINTER(c_size_t)]
    lib.vsl_frame_mmap.restype = c_void_p

    lib.vsl_frame_munmap.argtypes = [c_void_p]
    lib.vsl_frame_munmap.restype = None

    lib.vsl_frame_sync.argtypes = [c_void_p, c_int, c_int]
    lib.vsl_frame_sync.restype = c_int

    # Host API
    lib.vsl_host_init.argtypes = [c_char_p]
    lib.vsl_host_init.restype = c_void_p

    lib.vsl_host_release.argtypes = [c_void_p]
    lib.vsl_host_release.restype = None

    lib.vsl_host_path.argtypes = [c_void_p]
    lib.vsl_host_path.restype = c_char_p

    lib.vsl_host_poll.argtypes = [c_void_p, c_int64]
    lib.vsl_host_poll.restype = c_int

    lib.vsl_host_process.argtypes = [c_void_p]
    lib.vsl_host_process.restype = c_int

    lib.vsl_host_sockets.argtypes = [
        c_void_p, c_size_t, POINTER(c_int), POINTER(c_size_t)]
    lib.vsl_host_sockets.restype = c_int

    lib.vsl_host_post.argtypes = [
        c_void_p, c_void_p, c_int64, c_int64, c_int64, c_int64]
    lib.vsl_host_post.restype = c_int

    # Client API
    lib.vsl_client_init.argtypes = [c_char_p, c_void_p, c_bool]
    lib.vsl_client_init.restype = c_void_p

    lib.vsl_client_release.argtypes = [c_void_p]
    lib.vsl_client_release.restype = None

    lib.vsl_client_disconnect.argtypes = [c_void_p]
    lib.vsl_client_disconnect.restype = None

    lib.vsl_client_userptr.argtypes = [c_void_p]
    lib.vsl_client_userptr.restype = c_void_p

    lib.vsl_client_path.argtypes = [c_void_p]
    lib.vsl_client_path.restype = c_char_p

    lib.vsl_client_set_timeout.argtypes = [c_void_p, c_float]
    lib.vsl_client_set_timeout.restype = None

    lib.vsl_frame_wait.argtypes = [c_void_p, c_int64]
    lib.vsl_frame_wait.restype = c_void_p


if 'lib' not in locals():
    lib = load_library()
    load_symbols(lib)



# SPDX-License-Identifier: Apache-2.0
# Copyright Ⓒ 2025 Au-Zone Technologies. All Rights Reserved.

from ctypes import \
    byref, c_int, c_int64, c_size_t, c_void_p, py_object, pythonapi
from enum import IntEnum
from pathlib import Path
from typing import Optional, Union
from typeguard import typechecked
from .library import lib, errmsg
from .frame import Frame


@typechecked
class Client:
    """
    The VideoStream client connects to a host in order to receive frames which
    have been published on the host-side.  A client can connect to a single
    host but a host can have many client connections.

    Captured frames can be used as normal frames though with some semantic
    differences.  The underlying buffer should not be changed by calling alloc
    nor dealloc as doing so will release the frame back to the host and would
    be equivalent to creating a new frame with similar parameters but without
    the underlying buffer and therefore without the contents.

    Frames have a limited timespan controlled by the host, the default being
    100ms.  The local frame object remains valid beyond this time but attempts
    to lock or map after expiration will raise an error.  Locking a frame stops
    the frame from expiring and allows software which needs more time to read
    and process the frame to use as much time as they need.

    NOTE: Memory is a limited resource and this is especially true for DMA
    memory.  Locking too many frames will lead to exhausted resources.  Make
    sure to only lock as required and return frames once they are no longer
    needed!

    """

    def __init__(self, path: Union[str, Path], reconnect: bool = False):
        if isinstance(path, Path):
            path = path.as_posix()
        self._ptr = lib.vsl_client_init(path.encode('utf-8'), None, reconnect)
        if self._ptr is None:
            raise RuntimeError('Failed to create host: %s' % errmsg())

    def __del__(self):
        if hasattr(self, '_ptr') and self._ptr is not None:
            lib.vsl_client_release(self._ptr)
            self._ptr = None

    @property
    def path(self) -> str:
        p = lib.vsl_client_path(self._ptr)
        if p is None:
            raise RuntimeError('Failed to read host path: %s' % errmsg())
        return p.decode('utf-8')

    def disconnect(self):
        lib.vsl_client_disconnect(self._ptr)

    def frame_wait(self, until: int = 0) -> Frame:
        frame_ptr = lib.vsl_frame_wait(self._ptr, until)
        if frame_ptr is None:
            raise RuntimeError('Failed receive frame: %s' % errmsg())
        return Frame(0, 0, 0, ptr=frame_ptr)
